
import asyncio
import logging
from decimal import Decimal
from typing import Optional

from a_sync import a_sync
from brownie import chain

from y.contracts import Contract
from y.datatypes import Address, Block
from y.networks import Network
from y.prices import magic

logger = logging.getLogger(__name__)

KP3R = "0x1cEB5cB57C4D4E2b2433641b95Dd330A33185A44"
RKP3R = "0xEdB67Ee1B171c4eC66E6c10EC43EDBbA20FaE8e9"

def is_rkp3r(address: Address) -> bool:
    return chain.id == Network.Mainnet and address == RKP3R

@a_sync(default="sync")
async def get_price(address: Address, block: Optional[Block] = None) -> Decimal:
    price, discount = await asyncio.gather(magic.get_price(KP3R, block=block, sync=False), get_discount(block))
    return price * (100 - discount) / 100

async def get_discount(block: Optional[Block] = None) -> Decimal:
    rkp3r = await Contract.coroutine(RKP3R)
    discount = Decimal(await rkp3r.discount.coroutine(block_identifier=block))
    logger.debug("discount: %s", discount)
    return discount




import asyncio
import itertools
import logging
from contextlib import suppress
from decimal import Decimal
from typing import Any, AsyncIterator, Dict, List, Optional, Tuple

import a_sync
import brownie
from brownie import chain
from brownie.network.event import _EventItem
from eth_abi.exceptions import InsufficientDataBytes
from multicall import Call
from web3.exceptions import ContractLogicError

from y import convert
from y.classes.common import ERC20, ContractBase, WeiBalance
from y.constants import STABLECOINS, WRAPPED_GAS_COIN, sushi, usdc, weth
from y.contracts import Contract, contract_creation_block_async
from y.datatypes import (Address, AddressOrContract, AnyAddressType, Block,
                         Pool, UsdPrice)
from y.decorators import continue_on_revert, stuck_coro_debugger
from y.exceptions import (CantFindSwapPath, ContractNotVerified,
                          NonStandardERC20, NotAUniswapV2Pool, TokenNotFound,
                          call_reverted)
from y.interfaces.uniswap.factoryv2 import UNIV2_FACTORY_ABI
from y.networks import Network
from y.prices import magic
from y.prices.dex.uniswap.v2_forks import (ROUTER_TO_FACTORY,
                                           ROUTER_TO_PROTOCOL, special_paths)
from y.utils.dank_mids import dank_w3
from y.utils.events import ProcessedEvents
from y.utils.multicall import \
    multicall_same_func_same_contract_different_inputs
from y.utils.raw_calls import raw_call

logger = logging.getLogger(__name__)

Path = List[AddressOrContract]
Reserves = Tuple[int,int,int]

class UniswapV2Pool(ERC20):
    __slots__ = 'get_reserves', '_token0', '_token1', '_types_assumed'
    def __init__(
        self, 
        address: AnyAddressType, 
        token0: Optional[Address] = None, 
        token1: Optional[Address] = None, 
        deploy_block: Optional[int] = None, 
        asynchronous: bool = False,
    ):
        super().__init__(address, asynchronous=asynchronous)
        self.get_reserves = Call(self.address, 'getReserves()((uint112,uint112,uint32))')
        self._deploy_block = deploy_block
        self._token0 = token0
        self._token1 = token1
        self._types_assumed = True
        
    @a_sync.aka.cached_property
    async def factory(self) -> Address:
        try: return await raw_call(self.address, 'factory()', output='address', sync=False)
        except ValueError as e:
            if call_reverted(e):
                raise NotAUniswapV2Pool from e
            # `is not a valid ETH address` means we got some kind of response from the chain.
            # but couldn't convert to address. If it happens to be a goofy but
            # verified uni fork, maybe we can get factory this way
            okay_errors = ['is not a valid ETH address','invalid opcode','invalid jump destination']
            if any([msg in str(e) for msg in okay_errors]):
                contract = await Contract.coroutine(self.address)
                try: 
                    return await contract.factory.coroutine()
                except AttributeError as exc:
                    raise NotAUniswapV2Pool from exc
            else:
                raise

    @a_sync.aka.property
    async def tokens(self) -> Tuple[ERC20, ERC20]:
        return await asyncio.gather(self.__token0__(sync=False), self.__token1__(sync=False))
    
    @a_sync.aka.property
    async def token0(self) -> ERC20:
        if self._token0 is None:
            with suppress(ContractLogicError):
                if token0 := await Call(self.address, ['token0()(address)']).coroutine():
                    self._token0 = ERC20(token0, asynchronous=self.asynchronous)
        if self._token0 is None:
            raise NotAUniswapV2Pool(self.address)
        return self._token0

    @a_sync.aka.property
    async def token1(self) -> ERC20:
        if self._token1 is None:
            with suppress(ContractLogicError):
                if token1 := await Call(self.address, ['token1()(address)']).coroutine():
                    return ERC20(token1, asynchronous=self.asynchronous)
        if self._token1 is None:
            raise NotAUniswapV2Pool(self.address)
        return self._token1
    
    @a_sync.a_sync(cache_type='memory')
    @stuck_coro_debugger
    async def get_price(self, block: Optional[Block] = None) -> Optional[UsdPrice]:
        tvl = await self.tvl(block=block, sync=False)
        if tvl is not None:
            # TODO: move decimal conversion into total_supply_readable
            return UsdPrice(tvl / Decimal(await self.total_supply_readable(block=block, sync=False)))
        return None
    
    @stuck_coro_debugger
    async def reserves(self, block: Optional[Block] = None) -> Optional[Tuple[WeiBalance, WeiBalance]]:
        reserves, tokens = await asyncio.gather(self.get_reserves.coroutine(block_id=block), self.__tokens__(sync=False))

        if reserves is None and self._types_assumed:
            try:
                self._check_return_types()
            except AttributeError as e:
                raise NotAUniswapV2Pool(self.address) from e
            return await self.reserves(block, sync=False)
        
        if reserves is None and self._verified:
            # This shouldn't really run anymore, maybe delete
            try:
                reserves = await self.contract.getReserves.coroutine(block_identifier=block)
                types = ",".join(output["type"] for output in self.contract.getReserves.abi["outputs"])
                logger.warning(f'abi for getReserves for {self.contract} is {types}')
            except Exception as e:
                if not call_reverted(e):
                    raise e
                    
        if reserves is None:
            reserves = 0, 0

        return (WeiBalance(reserves[i], tokens[i], block=block) for i in range(2))

    @stuck_coro_debugger
    async def tvl(self, block: Optional[Block] = None) -> Optional[Decimal]:
        prices, reserves = await asyncio.gather(
            asyncio.gather(*[token.price(block=block, return_None_on_failure=True, sync=False) for token in await self.__tokens__(sync=False)]),
            self.reserves(block=block, sync=False),
        )

        if reserves is None:
            return None

        vals = [
            None if price is None else Decimal(await reserve.__readable__(sync=False)) * Decimal(price)
            for reserve, price in zip(reserves, prices)
        ]

        if vals[0] is not None and not vals[1]:
            vals[1] = vals[0]
        if vals[1] is not None and not vals[0]:
            vals[0] = vals[1]

        logger.debug('reserves: %s', reserves)
        logger.debug('prices: %s', prices)
        logger.debug('vals: %s', vals)
        if vals[0] is not None and vals[1] is not None:
            return sum(vals)
    
    @stuck_coro_debugger
    @a_sync.a_sync(ram_cache_maxsize=10_000, ram_cache_ttl=10*60)
    async def check_liquidity(self, token: Address, block: Block) -> int:
        if block and block < await self.deploy_block(sync=False):
            return 0
        try:
            if reserves := await self.reserves(block, sync=False):
                balance: WeiBalance
                for balance in reserves:
                    if token == balance.token:
                        return balance.balance
                raise TokenNotFound(f"{token} not found in {reserves}")
        except InsufficientDataBytes:
            return 0

    @stuck_coro_debugger
    async def is_uniswap_pool(self, block: Optional[Block] = None) -> bool:
        try:
            return all(await asyncio.gather(self.reserves(block, sync=False), self.total_supply(block, sync=False)))
        except (NotAUniswapV2Pool, InsufficientDataBytes):
            return False
        
    def _check_return_types(self) -> None:
        if not self._types_assumed:
            return
        try:
            reserves_types = ",".join(output["type"] for output in self.contract.getReserves.abi["outputs"])
            self._verified = True
            assert reserves_types.count(',') == 2, reserves_types
            self.get_reserves = Call(self.address, f'getReserves()(({reserves_types}))')
        except ContractNotVerified:
            self._verified = False
        self._types_assumed = False


class PoolsFromEvents(ProcessedEvents[UniswapV2Pool]):
    PairCreated = "0x0d3648bd0f6ba80134a33ba9275ac585d9d315f0ad8355cddefde31afa28d0e9"
    __slots__ = "asynchronous", "label"
    def __init__(self, factory: AnyAddressType, label: str, asynchronous: bool = False):
        self.asynchronous = asynchronous
        self.label = label
        super().__init__(addresses=[factory], topics=[[self.PairCreated]])
    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} label={self.label}>"
    async def pools(self, to_block: Optional[int] = None) -> AsyncIterator[UniswapV2Pool]:
        async for pool in self._objects_thru(block=to_block):
            yield pool
    def _get_block_for_obj(self, obj: UniswapV2Pool) -> int:
        return obj._deploy_block
    def _process_event(self, event: _EventItem) -> UniswapV2Pool:
        pool = UniswapV2Pool(
            address=event["pair"], 
            token0=event["token0"], 
            token1=event["token1"], 
            asynchronous=self.asynchronous,
        )
        # Do this here instead of in the init in case the user inited their own UniswapV2Pool object previoulsy, which is now the singleton
        pool._deploy_block = event.block_number
        return pool
    

class UniswapRouterV2(ContractBase):
    def __init__(self, router_address: AnyAddressType, *args: Any, **kwargs: Any) -> None:
        super().__init__(router_address, *args, **kwargs)

        self.label = ROUTER_TO_PROTOCOL[self.address]
        self.factory = ROUTER_TO_FACTORY[self.address]
        self.special_paths = special_paths(self.address)
        self.get_amounts_out = Call(self.address, "getAmountsOut(uint,address[])(uint[])")

        # we need the factory contract object cached in brownie so we can decode logs properly
        if not ContractBase(self.factory, asynchronous=self.asynchronous)._is_cached:
            brownie.Contract.from_abi('UniClone Factory [forced]', self.factory, UNIV2_FACTORY_ABI)

        self._events = PoolsFromEvents(self.factory, self.label, asynchronous=self.asynchronous)
        
    
    def __repr__(self) -> str:
        return f"<UniswapV2Router {self.label} '{self.address}'>"

    @stuck_coro_debugger
    @a_sync.a_sync(ram_cache_maxsize=500)
    async def get_price(
        self,
        token_in: Address,
        block: Optional[Block] = None,
        token_out: Address = usdc.address,
        paired_against: Address = WRAPPED_GAS_COIN,
        ignore_pools: Tuple[Pool, ...] = (),
        ) -> Optional[UsdPrice]:
        """
        Calculate a price based on Uniswap Router quote for selling one `token_in`.
        Always uses intermediate WETH pair if `[token_in,weth,token_out]` swap path available.
        """

        token_in, token_out, path = str(token_in), str(token_out), None

        if chain.id == Network.BinanceSmartChain and token_out == usdc.address:
            busd = await Contract.coroutine("0xe9e7CEA3DedcA5984780Bafc599bD69ADd087D56")
            token_out = busd.address

        if token_in in STABLECOINS:
            return 1

        try:
            amount_in = await ERC20(token_in, asynchronous=True).scale
        except NonStandardERC20:
            return None

        if token_in in [weth.address, WRAPPED_GAS_COIN] and token_out in STABLECOINS:
            path = [token_in, token_out]

        elif str(token_out) in STABLECOINS:
            with suppress(CantFindSwapPath):
                path = await self.get_path_to_stables(token_in, block, _ignore_pools=ignore_pools, sync=False)
                logger.debug('smrt')

        # If we can't find a good path to stables, we might still be able to determine price from price of paired token
        if path is None and (deepest_pool:= await self.deepest_pool(token_in, block, _ignore_pools=ignore_pools, sync=False)):
            logger.debug('deepest pool: %s', deepest_pool)
            paired_with = (await self.get_pools_for(token_in, sync=False))[deepest_pool]
            path = [token_in, paired_with]
            quote, out_scale = await asyncio.gather(self.get_quote(amount_in, path, block=block, sync=False), ERC20(path[-1], asynchronous=True).scale)
            logger.debug('quote: %s', quote)
            if quote is not None:
                amount_out = Decimal(quote[-1]) / out_scale  
                fees = Decimal(0.997 ** (len(path) - 1))
                amount_out /= fees
                paired_with_price = await magic.get_price(paired_with, block, fail_to_None=True, ignore_pools=(*ignore_pools, deepest_pool), sync=False)

                if paired_with_price:
                    return amount_out * Decimal(paired_with_price)

        # If we still don't have a workable path, try this smol brain method
        if path is None:
            path = self._smol_brain_path_selector(token_in, token_out, paired_against)
            logger.debug('smol')

        fees = 0.997 ** (len(path) - 1)
        logger.debug('router: %s     path: %s', self.label, path)
        quote, out_scale = await asyncio.gather(self.get_quote(amount_in, path, block=block, sync=False), ERC20(path[-1],asynchronous=True).scale)
        if quote is not None:
            amount_out = quote[-1] / out_scale
            return UsdPrice(amount_out / fees)


    @continue_on_revert
    @stuck_coro_debugger
    async def get_quote(self, amount_in: int, path: Path, block: Optional[Block] = None) -> Tuple[int,int]:
        if not self._is_cached:
            return await self.get_amounts_out.coroutine((amount_in, path), block_id=block)
        try:
            return await self.contract.getAmountsOut.coroutine(amount_in, path, block_identifier=block)
        # TODO figure out how to best handle uni forks with slight modifications.
        # Sometimes the below "else" code will not work with modified methods. Brownie works for now.
        except Exception as e:
            strings = [
                "INSUFFICIENT_INPUT_AMOUNT",
                "INSUFFICIENT_LIQUIDITY",
                "INSUFFICIENT_OUT_LIQUIDITY",
                "Sequence has incorrect length",
            ]
            if not call_reverted(e) and all(s not in str(e) for s in strings):
                raise e

    def _smol_brain_path_selector(self, token_in: AddressOrContract, token_out: AddressOrContract, paired_against: AddressOrContract) -> Path:
        '''Chooses swap path to use for quote'''
        token_in, token_out, paired_against = str(token_in), str(token_out), str(paired_against)

        if str(paired_against) in STABLECOINS and str(token_out) in STABLECOINS:            path = [token_in, paired_against]
        elif weth in (token_in, token_out):                                                 path = [token_in, token_out]
        elif sushi and paired_against == sushi and token_out != sushi:                      path = [token_in,sushi,weth,token_out]
        elif str(token_in) in self.special_paths and str(token_out) in STABLECOINS:         path = self.special_paths[str(token_in)]

        elif chain.id == Network.BinanceSmartChain:
            from y.constants import cake, wbnb
            if WRAPPED_GAS_COIN in (token_in, token_out):                                   path = [token_in, token_out]
            elif cake.address in (token_in, token_out):                                     path = [token_in, token_out]
            else:                                                                           path = [token_in,wbnb.address,token_out]
        else:
            if WRAPPED_GAS_COIN in (token_in, token_out):                                   path = [token_in, token_out]
            else:                                                                           path = [token_in, WRAPPED_GAS_COIN, token_out]

        return path
    
    @a_sync.aka.cached_property
    @stuck_coro_debugger
    async def pools(self) -> List[UniswapV2Pool]:
        logger.info('Fetching pools for %s on %s. If this is your first time using ypricemagic, this can take a while. Please wait patiently...', self.label, Network.printable())
        pools = [pool async for pool in self._events.pools(to_block=await dank_w3.eth.block_number)]
        all_pairs_len = await raw_call(self.factory, 'allPairsLength()', block=chain.height, output='int', sync=False)
        if len(pools) > all_pairs_len:
            raise NotImplementedError('this shouldnt happen again')
        elif len(pools) < all_pairs_len: # <
            to_get = all_pairs_len - len(pools)
            logger.debug("Oh no! Looks like your node can't look back that far. Checking for the missing %s pools...", to_get)
            for pool in await multicall_same_func_same_contract_different_inputs(self.factory, 'allPairs(uint256)(address)', inputs=range(to_get), sync=False):
                logger.debug('pool: %s', pool)
                pools.insert(0, UniswapV2Pool(address=pool, asynchronous=self.asynchronous))
        tokens = set(await asyncio.gather(*itertools.chain(*((pool.token0, pool.token1) for pool in pools))))
        logger.info('Loaded %s pools supporting %s tokens on %s', len(pools), len(tokens), self.label)
        return pools

    @stuck_coro_debugger
    @a_sync.a_sync(ram_cache_maxsize=None, ram_cache_ttl=1800)
    async def get_pools_for(self, token_in: Address) -> Dict[UniswapV2Pool, Address]:
        pool: UniswapV2Pool
        pool_to_token_out = {}
        for pool in await self.__pools__(sync=False):
            token0, token1 = await asyncio.gather(pool.__token0__(sync=False), pool.__token1__(sync=False))
            if token_in == token0:
                pool_to_token_out[pool] = token1
            elif token_in == token1:
                pool_to_token_out[pool] = token0
        return pool_to_token_out

    @stuck_coro_debugger
    async def pools_for_token(self, token_address: Address, block: Optional[Block] = None, _ignore_pools: Tuple[UniswapV2Pool,...] = ()) -> Dict[UniswapV2Pool, Address]:
        pools: Dict[UniswapV2Pool, Address]
        pools = await self.get_pools_for(token_address, sync=False)
        pools = {k: v for k, v in pools.items() if k not in _ignore_pools}
        if pools and block is not None:
            deploy_blocks = await asyncio.gather(*[pool.deploy_block(when_no_history_return_0=True, sync=False) for pool in pools])
            pools = {k: v for (k, v), deploy_block in zip(pools.items(), deploy_blocks) if deploy_block <= block}
        return pools

    @stuck_coro_debugger
    @a_sync.a_sync(ram_cache_maxsize=500)
    async def deepest_pool(self, token_address: AnyAddressType, block: Optional[Block] = None, _ignore_pools: Tuple[UniswapV2Pool,...] = ()) -> Optional[UniswapV2Pool]:
        token_address = convert.to_address(token_address)
        if token_address == WRAPPED_GAS_COIN or token_address in STABLECOINS:
            return await self.deepest_stable_pool(token_address, block, sync=False)
        pools: Dict[UniswapV2Pool, Address] = await self.pools_for_token(token_address, block, _ignore_pools=_ignore_pools, sync=False)
        if not pools:
            return None
        liquidity = await asyncio.gather(*[pool.check_liquidity(token_address, block, sync=False) for pool in pools])

        deepest_pool = None
        deepest_pool_balance = 0
        for pool, depth in zip(pools, liquidity):
            if depth and depth > deepest_pool_balance:
                deepest_pool = pool
                deepest_pool_balance = depth
        return deepest_pool

    @stuck_coro_debugger
    @a_sync.a_sync(ram_cache_maxsize=500)
    async def deepest_stable_pool(self, token_address: AnyAddressType, block: Optional[Block] = None, _ignore_pools: Tuple[UniswapV2Pool,...] = ()) -> Optional[UniswapV2Pool]:
        token_address = convert.to_address(token_address)
        pools: Dict[UniswapV2Pool, Address]
        pools = {
            pool: paired_with
            for pool, paired_with in (await self.pools_for_token(token_address, None, _ignore_pools=_ignore_pools, sync=False)).items()
            if paired_with in STABLECOINS
        }

        if pools and block is not None:
            deploy_blocks = await asyncio.gather(*[pool.deploy_block(when_no_history_return_0=True, sync=False) for pool in pools])
            pools = {pool: paired_with for (pool, paired_with), deploy_block in zip(pools.items(), deploy_blocks) if deploy_block <= block}
            
        if not pools:
            return None
        liquidity = await asyncio.gather(*[pool.check_liquidity(token_address, block, sync=False) for pool in pools])

        deepest_stable_pool = None
        deepest_stable_pool_balance = 0
        for pool, depth in zip(pools, liquidity):
            if depth > deepest_stable_pool_balance:
                deepest_stable_pool = pool
                deepest_stable_pool_balance = depth
        return deepest_stable_pool

    @stuck_coro_debugger
    @a_sync.a_sync(ram_cache_maxsize=500)
    async def get_path_to_stables(self, token: AnyAddressType, block: Optional[Block] = None, _loop_count: int = 0, _ignore_pools: Tuple[UniswapV2Pool,...] = ()) -> Path:
        if _loop_count > 10:
            raise CantFindSwapPath
        token_address = convert.to_address(token)
        path = [token_address]
        deepest_pool = await self.deepest_pool(token_address, block, _ignore_pools, sync=False)
        if deepest_pool:
            paired_with = (await self.get_pools_for(token_address, sync=False))[deepest_pool]
            deepest_stable_pool = await self.deepest_stable_pool(token_address, block, _ignore_pools=_ignore_pools, sync=False)
            if deepest_stable_pool and deepest_pool == deepest_stable_pool:
                last_step = (await self.get_pools_for(token_address, sync=False))[deepest_stable_pool]
                path.append(last_step)
                return path

            if path == [token_address]:
                with suppress(CantFindSwapPath):
                    path.extend(
                        await self.get_path_to_stables(
                            paired_with,
                            block=block, 
                            _loop_count=_loop_count+1, 
                            _ignore_pools=tuple(list(_ignore_pools) + [deepest_pool]),
                            sync=False,
                        )
                    )

        if path == [token_address]:
            raise CantFindSwapPath(f'Unable to find swap path for {token_address} on {Network.printable()}')

        return path

    @stuck_coro_debugger
    @a_sync.a_sync(ram_cache_maxsize=100_000, ram_cache_ttl=30*60)
    async def check_liquidity(self, token: Address, block: Block, ignore_pools = []) -> int:
        if block and block < await contract_creation_block_async(self.factory):
            return 0
        pools: Dict[UniswapV2Pool, Address] = await self.pools_for_token(token, block=block, _ignore_pools=ignore_pools, sync=False)
        return max(await asyncio.gather(*[pool.check_liquidity(token, block) for pool in pools])) if pools else 0




import asyncio
import logging
from typing import Optional

import a_sync

from y import convert
from y.classes.common import ERC20
from y.contracts import Contract, has_methods
from y.datatypes import AnyAddressType, Block, UsdPrice
from y.prices import magic

logger = logging.getLogger(__name__)


@a_sync.a_sync(default='sync', cache_type='memory')
async def is_mstable_feeder_pool(address: AnyAddressType) -> bool:
    return await has_methods(address, ('getPrice()((uint,uint))','mAsset()(address)'), sync=False)

@a_sync.a_sync(default='sync')
async def get_price(token: AnyAddressType, block: Optional[Block] = None) -> UsdPrice:
    address = convert.to_address(token)
    contract = await Contract.coroutine(address)
    ratio, masset, scale = await asyncio.gather(
        contract.getPrice.coroutine(block_identifier=block),
        contract.mAsset.coroutine(block_identifier=block),
        ERC20(address, asynchronous=True).scale,
    )
    ratio = ratio[0] / scale
    underlying_price = await magic.get_price(masset, block, sync=False)
    return UsdPrice(underlying_price * ratio)




import logging

from a_sync import a_sync
from brownie import chain
from pony.orm import db_session

from y._db.common import token_attr_threads
from y._db.entities import Block, Chain, insert, retry_locked

logger = logging.getLogger(__name__)

@a_sync(default='async', executor=token_attr_threads)
@db_session
@retry_locked
def get_chain() -> Chain:
    return Chain.get(id=chain.id) or insert(type=Chain, id=chain.id) or Chain.get(id=chain.id)

@a_sync(default='async', executor=token_attr_threads)
@db_session
@retry_locked
def get_block(number: int) -> Block:
    chain = get_chain(sync=True)
    if block := Block.get(chain=chain, number=number):
        return block
    return insert(type=Block, chain=chain, number=number) or get_block(number, sync=True)




from typed_envs import EnvVarFactory, create_env

_envs = EnvVarFactory("YPRICEMAGIC")

# TTL for various in-memory caches thruout the library
CACHE_TTL = _envs.create_env("CACHE_TTL", int, default=60*60, verbose=False)
CONTRACT_CACHE_TTL = _envs.create_env("CONTRACT_CACHE_TTL", int, default=int(CACHE_TTL), verbose=False)

GETLOGS_BATCH_SIZE = _envs.create_env("GETLOGS_BATCH_SIZE", int, default=0)
GETLOGS_DOP = _envs.create_env("GETLOGS_DOP", int, default=32)

DB_PROVIDER = _envs.create_env("DB_PROVIDER", str, default="sqlite", verbose=False)
DB_HOST = _envs.create_env("DB_HOST", str, default="", verbose=False)
DB_PORT = _envs.create_env("DB_PORT", str, default="", verbose=False)
DB_USER = _envs.create_env("DB_USER", str, default="", verbose=False)
DB_PASSWORD = _envs.create_env("DB_PASSWORD", str, default="", verbose=False)
DB_DATABASE = _envs.create_env("DB_DATABASE", str, default="ypricemagic", verbose=False)
SKIP_CACHE = _envs.create_env("SKIP_CACHE", bool, default=False, verbose=False)

# ypriceapi
SKIP_YPRICEAPI = create_env("SKIP_YPRICEAPI", bool, default=False, verbose=False)



import abc
import asyncio
import logging
import threading
from collections import Counter, defaultdict
from itertools import zip_longest
from typing import (TYPE_CHECKING, Any, AsyncGenerator, AsyncIterator,
                    Callable, Dict, Iterable, List, NoReturn, Optional,
                    TypeVar)

import a_sync
import eth_retry
from a_sync.iter import ASyncIterable
from a_sync.primitives.executor import _AsyncExecutorMixin
from async_property import async_property
from brownie import web3
from brownie.convert.datatypes import EthAddress
from brownie.network.event import EventDict, _decode_logs, _EventItem
from dank_mids.semaphores import BlockSemaphore
from eth_typing import ChecksumAddress
from toolz import groupby
from web3.middleware.filter import block_ranges
from web3.types import LogReceipt

from y import ENVIRONMENT_VARIABLES as ENVS
from y._db.common import Filter, _clean_addresses
from y.datatypes import Address, Block
from y.utils.cache import memory
from y.utils.dank_mids import dank_w3
from y.utils.middleware import BATCH_SIZE

if TYPE_CHECKING:
    from y._db.utils.logs import LogCache


T = TypeVar('T')

logger = logging.getLogger(__name__)


def decode_logs(logs: List[LogReceipt]) -> EventDict:
    """
    Decode logs to events and enrich them with additional info.
    """
    try:
        decoded = _decode_logs(logs)
    except:
        decoded = []
        for log in logs:
            try:
                # get some help for debugging
                decoded.extend(_decode_logs([log]))
            except Exception as e:
                raise (log, *e.args) from e

    for i, log in enumerate(logs):
        setattr(decoded[i], "block_number", log["blockNumber"])
        setattr(decoded[i], "transaction_hash", log["transactionHash"])
        setattr(decoded[i], "log_index", log["logIndex"])
    return decoded


@a_sync.a_sync(default='sync')
async def get_logs_asap(
    address: Optional[Address],
    topics: Optional[List[str]],
    from_block: Optional[Block] = None,
    to_block: Optional[Block] = None,
    verbose: int = 0
) -> List[Any]:

    if from_block is None:
        from y.contracts import contract_creation_block_async
        from_block = 0 if address is None else await contract_creation_block_async(address, True)
    if to_block is None:
        to_block = await dank_w3.eth.block_number

    ranges = list(block_ranges(from_block, to_block, BATCH_SIZE))
    if verbose > 0:
        logger.info('fetching %d batches', len(ranges))
    
    batches = await asyncio.gather(*[_get_logs_async(address, topics, start, end) for start, end in ranges])
    return [log for batch in batches for log in batch]


async def get_logs_asap_generator(
    address: Optional[Address],
    topics: Optional[List[str]] = None,
    from_block: Optional[Block] = None,
    to_block: Optional[Block] = None,
    chronological: bool = True,
    run_forever: bool = False,
    run_forever_interval: int = 60,
    verbose: int = 0
) -> AsyncGenerator[List[LogReceipt], None]:
    # NOTE: If you don't need the logs in order, you will get your logs faster if you set `chronological` to False.

    if from_block is None:
        from y.contracts import contract_creation_block_async
        if address is None:
            from_block = 0
        elif isinstance(address, Iterable) and not isinstance(address, str):
            from_block = min(await asyncio.gather(*[contract_creation_block_async(addr, True) for addr in address]))
        else:
            from_block = await contract_creation_block_async(address, True)
    if to_block is None:
        to_block = await dank_w3.eth.block_number
    elif run_forever:
        raise TypeError('`to_block` must be None if `run_forever` is True.')
    if from_block > to_block:
        raise ValueError(f"from_block must be <= to_block. You passed from_block: {from_block} to_block: {to_block}.")
    while True:
        ranges = list(block_ranges(from_block, to_block, BATCH_SIZE))
        if verbose > 0:
            logger.info('fetching %d batches', len(ranges))
        coros = [_get_logs_async(address, topics, start, end) for start, end in ranges]
        if chronological:
            yielded = 0
            done = {}
            async for i, result in a_sync.as_completed(dict(enumerate(coros)), aiter=True):
                done[i] = result
                for i in range(len(coros)):
                    if yielded > i:
                        continue
                    if i not in done:
                        break
                    yield done.pop(i)
                    yielded += 1
        else:
            for logs in asyncio.as_completed(coros, timeout=None):
                yield await logs
        if not run_forever:
            return
        
        await asyncio.sleep(run_forever_interval)
        
        # Find start and end block for next loop
        current_block = await dank_w3.eth.block_number
        while current_block <= to_block:
            await asyncio.sleep(run_forever_interval)
            current_block = await dank_w3.eth.block_number
        from_block = to_block + 1 if to_block + 1 <= current_block else current_block
        to_block = current_block


def logs_to_balance_checkpoints(logs) -> Dict[EthAddress,int]:
    """
    Convert Transfer logs to `{address: {from_block: balance}}` checkpoints.
    """
    balances = Counter()
    checkpoints = defaultdict(dict)
    for block, block_logs in groupby('blockNumber', logs).items():
        events = decode_logs(block_logs)
        for log in events:
            # ZERO_ADDRESS tracks -totalSupply
            sender, receiver, amount = log.values()  # there can be several different aliases
            balances[sender] -= amount
            checkpoints[sender][block] = balances[sender]
            balances[receiver] += amount
            checkpoints[receiver][block] = balances[receiver]
    return checkpoints


def checkpoints_to_weight(checkpoints, start_block: Block, end_block: Block) -> float:
    total = 0
    for a, b in zip_longest(list(checkpoints), list(checkpoints)[1:]):
        if a < start_block or a > end_block:
            continue
        b = min(b, end_block) if b else end_block
        total += checkpoints[a] * (b - a) / (end_block - start_block)
    return total

@a_sync.a_sync
def _get_logs(
    address: Optional[ChecksumAddress],
    topics: Optional[List[str]],
    start: Block,
    end: Block
    ) -> List[LogReceipt]:
    if end - start == BATCH_SIZE - 1:
        response = _get_logs_batch_cached(address, topics, start, end)
    else:
        response = _get_logs_no_cache(address, topics, start, end)
    for log in response:
        if address and log.address != address:
            ''' I have this due to a corrupt cache on my local box that I would prefer not to lose. '''
            ''' It will not impact your scripts. ''' 
            response.remove(log)
    return response

get_logs_semaphore = defaultdict(
    lambda: BlockSemaphore(
        ENVS.GETLOGS_DOP, 
        # We need to do this in case users use the sync api in a multithread context
        name="y.get_logs" + "" if threading.current_thread() == threading.main_thread() else f".{threading.current_thread()}",
    )
)

async def _get_logs_async(address, topics, start, end) -> List[LogReceipt]:
    async with get_logs_semaphore[asyncio.get_event_loop()][end]:
        return await _get_logs(address, topics, start, end, asynchronous=True)
    
async def _get_logs_async_no_cache(address, topics, start, end) -> List[LogReceipt]:
    try:
        if address is None:
            return await dank_w3.eth.get_logs({"topics": topics, "fromBlock": start, "toBlock": end})
        elif topics is None:
            return await dank_w3.eth.get_logs({"address": address, "fromBlock": start, "toBlock": end})
        else:
            return await dank_w3.eth.get_logs({"address": address, "topics": topics, "fromBlock": start, "toBlock": end})
    except Exception as e:
        errs = [
            "Service Unavailable for url:",
            "exceed maximum block range",
            "block range is too wide",
            "request timed out",
            "parse error",
        ]
        if all(err not in str(e) for err in errs):
            raise
        logger.debug('your node is having trouble, breaking batch in half')
        batch_size = (end - start + 1)
        if batch_size <= 2:
            # breaks the logic below and usually just succeeds on retry anyway
            return await _get_logs_async_no_cache(address, topics, start, end)
        half_of_batch = batch_size // 2
        batch1_end = start + half_of_batch
        batch2_start = batch1_end + 1
        batch1 = await _get_logs_async_no_cache(address, topics, start, batch1_end)
        batch2 = await _get_logs_async_no_cache(address, topics, batch2_start, end)
        response = batch1 + batch2
    return response

@eth_retry.auto_retry
def _get_logs_no_cache(
    address: Optional[ChecksumAddress],
    topics: Optional[List[str]],
    start: Block,
    end: Block
    ) -> List[LogReceipt]:
    logger.debug('fetching logs %s to %s', start, end)
    try:
        if address is None:
            response = web3.eth.get_logs({"topics": topics, "fromBlock": start, "toBlock": end})
        elif topics is None:
            response = web3.eth.get_logs({"address": address, "fromBlock": start, "toBlock": end})
        else:
            response = web3.eth.get_logs({"address": address, "topics": topics, "fromBlock": start, "toBlock": end})
    except Exception as e:
        errs = [
            "Service Unavailable for url:",
            "exceed maximum block range",
            "block range is too wide",
            "request timed out",
            "parse error",
        ]
        if any(err in str(e) for err in errs):
            logger.debug('your node is having trouble, breaking batch in half')
            batch_size = (end - start + 1)
            half_of_batch = batch_size // 2
            batch1_end = start + half_of_batch
            batch2_start = batch1_end + 1
            batch1 = _get_logs_no_cache(address, topics, start, batch1_end)
            batch2 = _get_logs_no_cache(address, topics, batch2_start, end)
            response = batch1 + batch2
        else:
            raise
    return response



##############################
# Leaving this here so as to not upset people's caches
##############################


@memory.cache()
def _get_logs_batch_cached(
    address: Optional[str],
    topics: Optional[List[str]],
    start: Block,
    end: Block
    ) -> List[LogReceipt]:
    return _get_logs_no_cache(address, topics, start, end)


class LogFilter(Filter[LogReceipt, "LogCache"]):
    __slots__ = 'addresses', 'topics', 'from_block'
    def __init__(
        self, 
        *, 
        addresses = [], 
        topics = [], 
        from_block: Optional[int] = None,
        chunk_size: int = BATCH_SIZE,
        chunks_per_batch: Optional[int] = None,
        semaphore: Optional[BlockSemaphore] = None,
        executor: Optional[_AsyncExecutorMixin] = None,
        is_reusable: bool = True,
        verbose: bool = False,
    ):
        self.addresses = _clean_addresses(addresses)
        self.topics = topics
        super().__init__(from_block, chunk_size=chunk_size, chunks_per_batch=chunks_per_batch, semaphore=semaphore, executor=executor, is_reusable=is_reusable, verbose=verbose)
    
    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} addresses={self.addresses} topics={self.topics}>"
    @property
    def cache(self) -> "LogCache":
        if self._cache is None:
            from y._db.utils.logs import LogCache
            self._cache = LogCache(self.addresses, self.topics)
        return self._cache
    
    @property
    def semaphore(self) -> BlockSemaphore:
        if self._semaphore is None:
            self._semaphore = get_logs_semaphore[asyncio.get_event_loop()]
        return self._semaphore

    def logs(self, to_block: Optional[int]) -> AsyncIterator[LogReceipt]:
        return self._objects_thru(block=to_block)

    @property
    def insert_to_db(self) -> Callable[[LogReceipt], None]:
        from y._db.utils.logs import insert_log
        return insert_log
    
    @async_property
    async def _from_block(self) -> int:
        if self.from_block is None:
            from y.contracts import contract_creation_block_async
            if self.addresses is None:
                self.from_block = 0
            elif isinstance(self.addresses, Iterable) and not isinstance(self.addresses, str):
                self.from_block = min(await asyncio.gather(*[contract_creation_block_async(addr, when_no_history_return_0=True) for addr in self.addresses]))
            else:
                self.from_block = await contract_creation_block_async(self.addresses, when_no_history_return_0=True)
        return self.from_block
    
    async def _fetch_range(self, range_start: int, range_end: int) -> List[LogReceipt]:
        return await _get_logs_async_no_cache(self.addresses, self.topics, range_start, range_end)

    async def _fetch(self) -> NoReturn:
        from_block = await self._from_block
        await self._loop(from_block)


class Events(LogFilter):
    obj_type = _EventItem
    __slots__ = []
    def events(self, to_block: int) -> AsyncIterator[_EventItem]:
        return self._objects_thru(block=to_block)
    def _extend(self, objs) -> None:
        return self._objects.extend(decode_logs(objs))
    def _get_block_for_obj(self, obj: _EventItem) -> int:
        return obj.block_number

class ProcessedEvents(Events, ASyncIterable[T]):
    __slots__ = []
    def _include_event(self, event: _EventItem) -> bool:
        """Override this to exclude specific events from processing and collection."""
        return True
    @abc.abstractmethod
    def _process_event(self, event: _EventItem) -> T:
        ...
    def objects(self, to_block: int) -> AsyncIterator[_EventItem]:
        return self._objects_thru(block=to_block)
    def _extend(self, logs: List[LogReceipt]) -> None:
        for event in decode_logs(logs):
            if self._include_event(event):
                self._objects.append(self._process_event(event))



import asyncio
import logging
from typing import Optional, Tuple

import a_sync
from brownie import ZERO_ADDRESS, chain

from y.classes.common import ERC20
from y.constants import usdc
from y.contracts import Contract, contract_creation_block_async
from y.datatypes import Address, Block, Pool, UsdPrice
from y.decorators import stuck_coro_debugger
from y.exceptions import UnsupportedNetwork, continue_if_call_reverted
from y.networks import Network
from y.utils.raw_calls import _decimals

logger = logging.getLogger(__name__)

class UniswapV1(a_sync.ASyncGenericBase):
    factory = "0xc0a47dFe034B400B47bDaD5FecDa2621de6c4d95"

    def __init__(self, asynchronous: bool = False) -> None:
        if chain.id != Network.Mainnet:
            raise UnsupportedNetwork(f"UniswapV1 does not suppport chainid {chain.id}")
        self.asynchronous = asynchronous
        
    @stuck_coro_debugger
    @a_sync.a_sync(ram_cache_maxsize=None)
    async def get_exchange(self, token_address: Address) -> Optional[Contract]:
        factory = await Contract.coroutine(self.factory)
        exchange = await factory.getExchange.coroutine(token_address)
        if exchange != ZERO_ADDRESS:
            return await Contract.coroutine(exchange)

    @stuck_coro_debugger
    async def get_price(
        self, 
        token_address: Address, 
        block: Optional[Block],
        ignore_pools: Tuple[Pool, ...] = (),
        ) -> Optional[UsdPrice]:
        exchange, usdc_exchange, decimals = await asyncio.gather(
            self.get_exchange(token_address, sync=False),
            self.get_exchange(usdc, sync=False),
            _decimals(token_address, block, sync=False),
        )
        if exchange is None:
            return None
        
        try:
            eth_bought = await exchange.getTokenToEthInputPrice.coroutine(10 ** decimals, block_identifier=block)
            usdc_bought = await usdc_exchange.getEthToTokenInputPrice.coroutine(eth_bought, block_identifier=block) / 1e6
            fees = 0.997 ** 2
            return UsdPrice(usdc_bought / fees)
        except ValueError as e:
            if 'invalid jump destination' in str(e):
                return None
            continue_if_call_reverted(e)

    @stuck_coro_debugger
    @a_sync.a_sync(ram_cache_maxsize=10_000, ram_cache_ttl=10*60)
    async def check_liquidity(self, token_address: Address, block: Block, ignore_pools: Tuple[Pool, ...] = ()) -> int:
        exchange = await self.get_exchange(token_address, sync=False)
        if exchange is None or exchange in ignore_pools:
            return 0
        if block < await contract_creation_block_async(exchange):
            return 0
        return await ERC20(token_address, asynchronous=True).balance_of(exchange, block)




import asyncio
from decimal import Decimal
from typing import List, Optional

from a_sync import a_sync
from pony.orm import TransactionIntegrityError, commit, db_session

from y import constants
from y._db.common import token_attr_threads
from y._db.entities import Price, retry_locked, insert
from y._db.utils._ep import _get_get_block, _get_get_token


@a_sync(default='async', executor=token_attr_threads)
@db_session
@retry_locked
def get_price(address: str, block: int) -> Optional[Decimal]:
    get_block = _get_get_block()
    get_token = _get_get_token()
    if address == constants.EEE_ADDRESS:
        address = constants.WRAPPED_GAS_COIN
    if price := Price.get(
        token = get_token(address, sync=True), 
        block = get_block(block, sync=True), 
    ):
        return price.price

async def set_price(address: str, block: int, price: Decimal) -> None:
    _tasks.append(asyncio.create_task(coro=_set_price(address, block, price), name=f"set_price {price} for {address} at {block}"))
    for t in _tasks[:]:
        if t.done():
            await t
            _tasks.remove(t)

@a_sync(default='async', executor=token_attr_threads)
@db_session
@retry_locked
def _set_price(address: str, block: int, price: Decimal) -> None:
    get_block = _get_get_block()
    get_token = _get_get_token()
    if address == constants.EEE_ADDRESS:
        address = constants.WRAPPED_GAS_COIN
    insert(
        type = Price,
        block = get_block(block, sync=True),
        token = get_token(address, sync=True),
        price = Decimal(price),
    )
    
_tasks: List[asyncio.Task] = []
